#include <u.h>
#include <libc.h>
#include <bio.h>
#include <ctype.h>
#include "ical.h"

static long
iso2sec(char *s, char *timezone)
{
	Tm tm;
	long t;
	char yer[5], mon[3], day[3], hrs[3], min[3], sec[32];
	char *td[4], *ymd[5], *hms[5], *tz;

	/* e.g.
	 * 2010-04-06T11:39:46Z
	 * 2011-11-27t00:00:00.000+0000
	 * 20111127t000000Z
	 */
	memset(&tm, 0, sizeof(tm));
	if(s == nil || *s == 0)
		return -1;

	if(gettokens(s, td, nelem(td), "tT") != 2)
		return -1L;

	if(gettokens(td[0], ymd, nelem(ymd), "-") != 3){
		ymd[0] = yer; strecpy(yer, yer+5, td[0]);
		ymd[1] = mon; strecpy(mon, mon+3, td[0]+4);
		ymd[2] = day; strecpy(day, day+3, td[0]+6);
	}

	if(gettokens(td[1], hms, nelem(hms), ":") != 2){
		hms[0] = hrs; strecpy(hrs, hrs+3, td[1]);
		hms[1] = min; strecpy(min, min+3, td[1]+2);
		hms[2] = sec; strecpy(sec, sec+32, td[1]+4);
	}

	tm.year = strtol(ymd[0], nil, 10) - 1900;
	tm.mon = strtol(ymd[1], nil, 10) -1;
	tm.mday = strtol(ymd[2], nil, 10);
	tm.hour = strtol(hms[0], nil, 10);
	tm.min = strtol(hms[1], nil, 10);
	tm.sec = strtol(hms[2], &tz, 10);

	if(*tz == '.')
		strtol(tz+1, &tz, 10);	/* skip fractional secconds */

	switch(*tz){
	case '-':		/* numeric offset */
	case '+':
		tm.tzoff = strtol(tz, nil, 10);
		break;
	case 'z':
	case 'Z':		/* Zulu time == UTC */
		break;
	default:		/* named zones not supported yet */
		if(timezone && strcmp(timezone, "GMT Standard Time") != 0)
			fprint(2, "%s: '%s' found, only GMT supported\n", argv0, timezone);
		break;
	}

	if((t = tm2sec(&tm)) < 0)
		return -1L;
	return t;
}

/* strip silly preludes on some Naim employee's names */
static char *
trimname(char *name)
{
	char *p;
	static char buf[64];

	if(name == nil)
		return nil;

	strecpy(buf, buf+sizeof(buf), name);
	/* strip silly trailing company / department designation */
	if((p = strstr(buf, " - Naim")) != nil)
		*p = 0;

	/* somtimes the format is sligtly different */
	if((p = strstr(buf, "- Naim")) != nil)
		*p = 0;

	return buf;
}

/* append to a malloc'ed string */
static char *
growstr(char **s, char *t)
{
	assert(t != nil);
	if(*s){
		*s = realloc(*s, strlen(*s) +2 +strlen(t) +1);
		assert(*s != nil);
		strcat(*s, ", ");
		strcat(*s, t);
	}
	else{
		*s = strdup(t);
		assert(*s != nil);
	}
	return *s;
}

/* print human friendly start and duration times */
static char *
timestr(long start, long finish)
{
	Tm *tm;
	char *ap;
	int n, hrs;
	long duration;
	static char buf[64];
	static char *day[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	static char *mon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

	if((tm = gmtime(start)) == nil)
		return "<gmtime failed>";

	snprint(buf, sizeof(buf), "%s %2d %s %d  ", day[tm->wday], tm->mday, mon[tm->mon], tm->year+1900);

	ap = "am";
	hrs = tm->hour;
	if(tm->hour > 12){
		ap = "pm";
		hrs -= 12;
	}

	n = strlen(buf);
	if(tm->min == 0)
		snprint(buf+n, sizeof(buf)-n, "%d%s	", hrs, ap);
	else
		snprint(buf+n, sizeof(buf)-n, "%d:%02d%s	", hrs, tm->min, ap);

	n = strlen(buf);
	duration = (finish - start)/60;
	if(duration == 0)
		snprint(buf+n, sizeof(buf)-n, "δ");
	else
	if(duration == 15)
		snprint(buf+n, sizeof(buf)-n, "¼h");
	else
	if(duration == 30)
		snprint(buf+n, sizeof(buf)-n, "½h");
	else
	if(duration == 45)
		snprint(buf+n, sizeof(buf)-n, "¾h");
	else
	if(duration < 60)
		snprint(buf+n, sizeof(buf)-n, "%ldm", duration);
	else
	if(duration >= 60*24 && (duration % 60*24) == 0)
		snprint(buf+n, sizeof(buf)-n, "%ldd", duration / (60*24));
	else
	if((duration % 60) == 0)
		snprint(buf+n, sizeof(buf)-n, "%ldh", duration/60);
	else
	if((duration % 60) == 15)
		snprint(buf+n, sizeof(buf)-n, "%ld¼h", duration/60);
	else
	if((duration % 60) == 30)
		snprint(buf+n, sizeof(buf)-n, "%ld½h", duration/60);
	else
	if((duration % 60) == 45)
		snprint(buf+n, sizeof(buf)-n, "%ld¾h", duration/60);
	else
		snprint(buf+n, sizeof(buf)-n, "%ld:%-ldh", duration/60, duration%60);

	return buf;
}

static Param *
lookparam(Param *root, char *name)
{
	Param *p;

	for(p = root; p; p = p->next)
		if(cistrcmp(p->str, name) == 0)
			return p;
	return nil;
}

static Attr *
lookattr(Attr *root, char *name)
{
	Attr *a;

	for(a = root; a; a = a->next)
		if(cistrcmp(a->name, name) == 0)
			return a;
	return nil;
}

static Value *
lookvalue(Value *root, char *name, int len)
{
	Value *v;

	for(v = root; v; v = v->next)
		if(cistrncmp(v->str, name, len) == 0)
			return v;
	return nil;
}

/* print events in calandar(1) format */
void
calendar(Biobuf *bp, Param *root)
{
	Param *p;
	Attr *a, *tz;
	long start, finish;
	char *attendees, *location, *summary, *organiser;

	start = finish = 0;
	summary = location = attendees = organiser = nil;
	for(p = root; p; p = p->next){
		if(cistrcmp(p->str, "begin") == 0 && p->values && cistrcmp(p->values->str, "vevent") == 0){
			free(attendees);
			start = finish = 0;
			summary = location = attendees = organiser = nil;
		}

		if(cistrcmp(p->str, "organizer") == 0)
			if((a = lookattr(p->attrs, "cn")) != nil)
				organiser = a->value;

		if(cistrcmp(p->str, "summary") == 0 && p->values)
			summary = p->values->str;

		if(cistrcmp(p->str, "location") == 0 && p->values)
			location = p->values->str;

		if(cistrcmp(p->str, "dtstart") == 0 && p->values)
			if((tz = lookattr(p->attrs, "tzid")) != nil)
				start = iso2sec(p->values->str, tz->value);
			else
				start = iso2sec(p->values->str, nil);

		if(cistrcmp(p->str, "dtend") == 0 && p->values)
			if((tz = lookattr(p->attrs, "tzid")) != nil)
				finish = iso2sec(p->values->str, tz->value);
			else
				finish = iso2sec(p->values->str, nil);

		if(cistrcmp(p->str, "attendee") == 0)
			if((a = lookattr(p->attrs, "cn")) != nil)
				growstr(&attendees, trimname(a->value));

		if(cistrcmp(p->str, "end") == 0 && p->values && cistrcmp(p->values->str, "vevent") == 0){
			Bprint(bp, "%s\t%s\t", timestr(start, finish), summary);
			if(location)
				Bprint(bp, "@ %s, ", location);
			if(organiser)
				Bprint(bp, " %s:", trimname(organiser));
			if(attendees)
				Bprint(bp, " %s", attendees);
			Bprint(bp, "\n");

			free(attendees);
			start = finish = 0;
			summary = location = attendees = organiser = nil;
		}
	}
}
